from typing import Any, Callable, List, Optional

from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputType,
    GraphQLOutputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLResolveInfo,
    get_nullable_type,
)

__all__ = ["plural_identifying_root_field"]


def plural_identifying_root_field(
    arg_name: str,
    input_type: GraphQLInputType,
    output_type: GraphQLOutputType,
    resolve_single_input: Callable[[GraphQLResolveInfo, str], Any],
    description: Optional[str] = None,
) -> GraphQLField:
    def resolve(_obj: Any, info: GraphQLResolveInfo, **args: Any) -> List:
        inputs = args[arg_name]
        return [resolve_single_input(info, input_) for input_ in inputs]

    return GraphQLField(
        GraphQLList(output_type),
        description=description,
        args={
            arg_name: GraphQLArgument(
                GraphQLNonNull(
                    GraphQLList(
                        GraphQLNonNull(get_nullable_type(input_type))  # type: ignore
                    )
                )
            )
        },
        resolve=resolve,
    )



from typing import NamedTuple

from pytest import mark

from graphql import (
    graphql,
    GraphQLField,
    GraphQLID,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)

from graphql_relay import node_definitions


class User(NamedTuple):
    id: str
    name: str


user_data = [User(id="1", name="John Doe"), User(id="2", name="Jane Smith")]

user_type: GraphQLObjectType

node_interface, node_field = node_definitions(
    lambda id_, _info: next(filter(lambda obj: obj.id == id_, user_data), None),
    lambda _obj, _info, _type: user_type.name,
)[:2]


user_type = GraphQLObjectType(
    "User",
    lambda: {
        "id": GraphQLField(GraphQLNonNull(GraphQLID)),
        "name": GraphQLField(GraphQLString),
    },
    interfaces=[node_interface],
)

query_type = GraphQLObjectType("Query", lambda: {"node": node_field})

schema = GraphQLSchema(query=query_type, types=[user_type])


def describe_node_interface_and_fields_with_async_object_fetcher():
    @mark.asyncio
    async def gets_the_correct_id_for_users():
        source = """
          {
            node(id: "1") {
              id
            }
          }
        """
        assert await graphql(schema, source) == ({"node": {"id": "1"}}, None)

    @mark.asyncio
    async def gets_the_correct_name_for_users():
        source = """
          {
            node(id: "1") {
              id
              ... on User {
                name
              }
            }
          }
        """
        assert await graphql(schema, source) == (
            {"node": {"id": "1", "name": "John Doe"}},
            None,
        )



"""This defines a basic set of data for our Star Wars Schema.

This data is hard coded for the sake of the demo, but you could imagine
fetching this data from a backend service rather than from hardcoded
JSON objects in a more complex demo.
"""

from typing import List, NamedTuple, Optional


class Ship(NamedTuple):
    id: str
    name: str


all_ships = [
    Ship("1", "X-Wing"),
    Ship("2", "Y-Wing"),
    Ship("3", "A-Wing"),
    # Yeah, technically it's Corellian. But it flew in the service of the rebels,
    # so for the purposes of this demo it's a rebel ship.
    Ship("4", "Millennium Falcon"),
    Ship("5", "Home One"),
    Ship("6", "TIE Fighter"),
    Ship("7", "TIE Interceptor"),
    Ship("8", "Executor"),
]


class Faction(NamedTuple):
    id: str
    name: str
    ships: List[str]


rebels = Faction("1", "Alliance to Restore the Republic", ["1", "2", "3", "4", "5"])

empire = Faction("2", "Galactic Empire", ["6", "7", "8"])

all_factions = [rebels, empire]


def create_ship(ship_name: str, faction_id: str) -> Ship:
    new_ship = Ship(str(len(all_ships) + 1), ship_name)
    all_ships.append(new_ship)
    faction = get_faction(faction_id)
    if faction:  # pragma: no cover else
        faction.ships.append(new_ship.id)
    return new_ship


def get_ship(id_: str) -> Optional[Ship]:
    return next(filter(lambda ship: ship.id == id_, all_ships), None)  # type: ignore


def get_faction(id_: str) -> Optional[Faction]:
    return next(
        filter(lambda faction: faction.id == id_, all_factions), None  # type: ignore
    )


def get_rebels() -> Faction:
    return rebels


def get_empire() -> Faction:
    return empire



"""The graphql_relay package"""

# The graphql-relay and graphql-relay-js version info
from .version import version, version_info, version_js, version_info_js

# Types and helpers for creating connection types in the schema
from .connection.connection import (
    backward_connection_args,
    connection_args,
    connection_definitions,
    forward_connection_args,
    page_info_type,
    Connection,
    ConnectionArguments,
    ConnectionConstructor,
    ConnectionCursor,
    ConnectionType,
    Edge,
    EdgeConstructor,
    EdgeType,
    GraphQLConnectionDefinitions,
    PageInfo,
    PageInfoConstructor,
    PageInfoType,
)

# Helpers for creating connections from arrays
from .connection.array_connection import (
    connection_from_array,
    connection_from_array_slice,
    cursor_for_object_in_connection,
    cursor_to_offset,
    get_offset_with_default,
    offset_to_cursor,
    SizedSliceable,
)

# Helper for creating mutations with client mutation IDs
from .mutation.mutation import (
    mutation_with_client_mutation_id,
    MutationFn,
    MutationFnWithoutArgs,
    NullResult,
)

# Helper for creating node definitions
from .node.node import node_definitions, GraphQLNodeDefinitions

#  Helper for creating plural identifying root fields
from .node.plural import plural_identifying_root_field

# Utilities for creating global IDs in systems that don't have them
from .node.node import from_global_id, global_id_field, to_global_id, ResolvedGlobalId

__version__ = version
__version_info__ = version_info
__version_js__ = version_js
__version_info_js__ = version_info_js

__all__ = [
    "backward_connection_args",
    "Connection",
    "ConnectionArguments",
    "ConnectionConstructor",
    "ConnectionCursor",
    "ConnectionType",
    "connection_args",
    "connection_from_array",
    "connection_from_array_slice",
    "connection_definitions",
    "cursor_for_object_in_connection",
    "cursor_to_offset",
    "Edge",
    "EdgeConstructor",
    "EdgeType",
    "forward_connection_args",
    "from_global_id",
    "get_offset_with_default",
    "global_id_field",
    "GraphQLConnectionDefinitions",
    "GraphQLNodeDefinitions",
    "MutationFn",
    "MutationFnWithoutArgs",
    "mutation_with_client_mutation_id",
    "node_definitions",
    "NullResult",
    "offset_to_cursor",
    "PageInfo",
    "PageInfoConstructor",
    "PageInfoType",
    "page_info_type",
    "plural_identifying_root_field",
    "ResolvedGlobalId",
    "SizedSliceable",
    "to_global_id",
    "version",
    "version_info",
    "version_js",
    "version_info_js",
]



from collections.abc import Mapping
from inspect import iscoroutinefunction
from typing import Any, Callable, Dict, Optional

from graphql import (
    resolve_thunk,
    GraphQLArgument,
    GraphQLField,
    GraphQLFieldMap,
    GraphQLInputField,
    GraphQLInputFieldMap,
    GraphQLInputObjectType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLString,
    ThunkMapping,
)
from graphql.pyutils import AwaitableOrValue

__all__ = [
    "mutation_with_client_mutation_id",
    "MutationFn",
    "MutationFnWithoutArgs",
    "NullResult",
]

# Note: Contrary to the Javascript implementation of MutationFn,
# the context is passed as part of the GraphQLResolveInfo and any arguments
# are passed individually as keyword arguments.
MutationFnWithoutArgs = Callable[[GraphQLResolveInfo], AwaitableOrValue[Any]]
# Unfortunately there is currently no syntax to indicate optional or keyword
# arguments in Python, so we also allow any other Callable as a workaround:
MutationFn = Callable[..., AwaitableOrValue[Any]]


class NullResult:
    def __init__(self, clientMutationId: Optional[str] = None) -> None:
        self.clientMutationId = clientMutationId


def mutation_with_client_mutation_id(
    name: str,
    input_fields: ThunkMapping[GraphQLInputField],
    output_fields: ThunkMapping[GraphQLField],
    mutate_and_get_payload: MutationFn,
    description: Optional[str] = None,
    deprecation_reason: Optional[str] = None,
    extensions: Optional[Dict[str, Any]] = None,
) -> GraphQLField:
    """
    Returns a GraphQLFieldConfig for the specified mutation.

    The input_fields and output_fields should not include `clientMutationId`,
    as this will be provided automatically.

    An input object will be created containing the input fields, and an
    object will be created containing the output fields.

    mutate_and_get_payload will receive a GraphQLResolveInfo as first argument,
    and the input fields as keyword arguments, and it should return an object
    (or a dict) with an attribute (or a key) for each output field.
    It may return synchronously or asynchronously.
    """

    def augmented_input_fields() -> GraphQLInputFieldMap:
        return dict(
            resolve_thunk(input_fields),
            clientMutationId=GraphQLInputField(GraphQLString),
        )

    def augmented_output_fields() -> GraphQLFieldMap:
        return dict(
            resolve_thunk(output_fields),
            clientMutationId=GraphQLField(GraphQLString),
        )

    output_type = GraphQLObjectType(name + "Payload", fields=augmented_output_fields)

    input_type = GraphQLInputObjectType(name + "Input", fields=augmented_input_fields)

    if iscoroutinefunction(mutate_and_get_payload):

        # noinspection PyShadowingBuiltins
        async def resolve(_root: Any, info: GraphQLResolveInfo, input: Dict) -> Any:
            payload = await mutate_and_get_payload(info, **input)
            clientMutationId = input.get("clientMutationId")
            if payload is None:
                return NullResult(clientMutationId)
            if isinstance(payload, Mapping):
                payload["clientMutationId"] = clientMutationId  # type: ignore
            else:
                payload.clientMutationId = clientMutationId
            return payload

    else:

        # noinspection PyShadowingBuiltins
        def resolve(  # type: ignore
            _root: Any, info: GraphQLResolveInfo, input: Dict
        ) -> Any:
            payload = mutate_and_get_payload(info, **input)
            clientMutationId = input.get("clientMutationId")
            if payload is None:
                return NullResult(clientMutationId)
            if isinstance(payload, Mapping):
                payload["clientMutationId"] = clientMutationId  # type: ignore
            else:
                payload.clientMutationId = clientMutationId  # type: ignore
            return payload

    return GraphQLField(
        output_type,
        description=description,
        deprecation_reason=deprecation_reason,
        args={"input": GraphQLArgument(GraphQLNonNull(input_type))},
        resolve=resolve,
        extensions=extensions,
    )



from typing import cast, Sequence

from pytest import deprecated_call, raises

from graphql_relay import (
    connection_from_array,
    connection_from_array_slice,
    cursor_for_object_in_connection,
    offset_to_cursor,
    Connection,
    Edge,
    PageInfo,
)

array_abcde = ["A", "B", "C", "D", "E"]

cursor_a = "YXJyYXljb25uZWN0aW9uOjA="
cursor_b = "YXJyYXljb25uZWN0aW9uOjE="
cursor_c = "YXJyYXljb25uZWN0aW9uOjI="
cursor_d = "YXJyYXljb25uZWN0aW9uOjM="
cursor_e = "YXJyYXljb25uZWN0aW9uOjQ="

edge_a = Edge(node="A", cursor=cursor_a)
edge_b = Edge(node="B", cursor=cursor_b)
edge_c = Edge(node="C", cursor=cursor_c)
edge_d = Edge(node="D", cursor=cursor_d)
edge_e = Edge(node="E", cursor=cursor_e)


def describe_connection_from_array():
    def warns_for_deprecated_import():
        from importlib import reload

        with deprecated_call():
            from graphql_relay.connection import arrayconnection as deprecated

            # noinspection PyDeprecation
            reload(deprecated)
        # noinspection PyDeprecation
        assert deprecated.connection_from_array is connection_from_array

    def describe_basic_slicing():
        def returns_all_elements_without_filters():
            c = connection_from_array(array_abcde, {})
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_a_smaller_first():
            c = connection_from_array(array_abcde, dict(first=2))
            assert c == Connection(
                edges=[
                    edge_a,
                    edge_b,
                ],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_b,
                    hasPreviousPage=False,
                    hasNextPage=True,
                ),
            )

        def respects_an_overly_large_first():
            c = connection_from_array(array_abcde, dict(first=10))
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_a_smaller_last():
            c = connection_from_array(array_abcde, dict(last=2))
            assert c == Connection(
                edges=[edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_d,
                    endCursor=cursor_e,
                    hasPreviousPage=True,
                    hasNextPage=False,
                ),
            )

        def respects_an_overly_large_last():
            c = connection_from_array(array_abcde, dict(last=10))
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

    def describe_pagination():
        def respects_first_and_after():
            c = connection_from_array(array_abcde, dict(first=2, after=cursor_b))
            assert c == Connection(
                edges=[edge_c, edge_d],
                pageInfo=PageInfo(
                    startCursor=cursor_c,
                    endCursor=cursor_d,
                    hasPreviousPage=False,
                    hasNextPage=True,
                ),
            )

        def respects_first_and_after_with_long_first():
            c = connection_from_array(array_abcde, dict(first=10, after=cursor_b))
            assert c == Connection(
                edges=[edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_c,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_last_and_before():
            c = connection_from_array(array_abcde, dict(last=2, before=cursor_d))
            assert c == Connection(
                edges=[edge_b, edge_c],
                pageInfo=PageInfo(
                    startCursor=cursor_b,
                    endCursor=cursor_c,
                    hasPreviousPage=True,
                    hasNextPage=False,
                ),
            )

        def respects_last_and_before_with_long_last():
            c = connection_from_array(array_abcde, dict(last=10, before=cursor_d))
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_c,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_first_and_after_and_before_too_few():
            c = connection_from_array(
                array_abcde,
                dict(first=2, after=cursor_a, before=cursor_e),
            )
            assert c == Connection(
                edges=[edge_b, edge_c],
                pageInfo=PageInfo(
                    startCursor=cursor_b,
                    endCursor=cursor_c,
                    hasPreviousPage=False,
                    hasNextPage=True,
                ),
            )

        def respects_first_and_after_and_before_too_many():
            c = connection_from_array(
                array_abcde,
                dict(first=4, after=cursor_a, before=cursor_e),
            )
            assert c == Connection(
                edges=[edge_b, edge_c, edge_d],
                pageInfo=PageInfo(
                    startCursor=cursor_b,
                    endCursor=cursor_d,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_first_and_after_and_before_exactly_right():
            c = connection_from_array(
                array_abcde,
                dict(first=3, after=cursor_a, before=cursor_e),
            )
            assert c == Connection(
                edges=[edge_b, edge_c, edge_d],
                pageInfo=PageInfo(
                    startCursor=cursor_b,
                    endCursor=cursor_d,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_last_and_after_and_before_too_few():
            c = connection_from_array(
                array_abcde,
                dict(last=2, after=cursor_a, before=cursor_e),
            )
            assert c == Connection(
                edges=[edge_c, edge_d],
                pageInfo=PageInfo(
                    startCursor=cursor_c,
                    endCursor=cursor_d,
                    hasPreviousPage=True,
                    hasNextPage=False,
                ),
            )

        def respects_last_and_after_and_before_too_many():
            c = connection_from_array(
                array_abcde,
                dict(last=4, after=cursor_a, before=cursor_e),
            )
            assert c == Connection(
                edges=[edge_b, edge_c, edge_d],
                pageInfo=PageInfo(
                    startCursor=cursor_b,
                    endCursor=cursor_d,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def respects_last_and_after_and_before_exactly_right():
            c = connection_from_array(
                array_abcde,
                dict(last=3, after=cursor_a, before=cursor_e),
            )
            assert c == Connection(
                edges=[edge_b, edge_c, edge_d],
                pageInfo=PageInfo(
                    startCursor=cursor_b,
                    endCursor=cursor_d,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

    def describe_cursor_edge_cases():
        def throws_an_error_if_first_smaller_than_zero():
            with raises(ValueError) as exc_info:
                connection_from_array(array_abcde, dict(first=-1))
            assert str(exc_info.value) == (
                "Argument 'first' must be a non-negative integer."
            )

        def throws_an_error_if_last_smaller_than_zero():
            with raises(ValueError) as exc_info:
                connection_from_array(array_abcde, dict(last=-1))
            assert str(exc_info.value) == (
                "Argument 'last' must be a non-negative integer."
            )

        def returns_all_elements_if_cursors_are_invalid():
            c1 = connection_from_array(
                array_abcde, dict(before="InvalidBase64", after="InvalidBase64")
            )

            invalid_unicode_in_base64 = "9JCAgA=="  # U+110000
            c2 = connection_from_array(
                array_abcde,
                dict(before=invalid_unicode_in_base64, after=invalid_unicode_in_base64),
            )

            assert c1 == c2
            assert c1 == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def returns_all_elements_if_cursors_are_on_the_outside():
            all_edges = Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

            assert (
                connection_from_array(array_abcde, dict(before=offset_to_cursor(6)))
                == all_edges
            )
            assert (
                connection_from_array(array_abcde, dict(before=offset_to_cursor(-1)))
                == all_edges
            )
            assert (
                connection_from_array(array_abcde, dict(after=offset_to_cursor(6)))
                == all_edges
            )
            assert (
                connection_from_array(array_abcde, dict(after=offset_to_cursor(-1)))
                == all_edges
            )

        def returns_no_elements_if_cursors_cross():
            c = connection_from_array(
                array_abcde,
                dict(before=cursor_c, after=cursor_e),
            )
            assert c == Connection(
                edges=[],
                pageInfo=PageInfo(
                    startCursor=None,
                    endCursor=None,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

    def describe_cursor_for_object_in_connection():
        def returns_an_edges_cursor_given_an_array_and_a_member_object():
            letter_b_cursor = cursor_for_object_in_connection(array_abcde, "B")
            assert letter_b_cursor == cursor_b

        def returns_null_given_an_array_and_a_non_member_object():
            letter_f_cursor = cursor_for_object_in_connection(array_abcde, "F")
            assert letter_f_cursor is None

        def describe_extended_functionality():
            """Test functionality that is not part of graphql-relay-js."""

            def returns_an_edges_cursor_given_an_array_without_index_method():
                class LettersWithoutIndex:
                    __getitem__ = array_abcde.__getitem__

                letters_without_index = cast(Sequence, LettersWithoutIndex())

                with raises(AttributeError):
                    letters_without_index.index("B")

                letter_b_cursor = cursor_for_object_in_connection(
                    letters_without_index, "B"
                )
                assert letter_b_cursor == cursor_b

                no_letter_cursor = cursor_for_object_in_connection(
                    letters_without_index, "="
                )
                assert no_letter_cursor is None

    def describe_extended_functionality():
        """Test functionality that is not part of graphql-relay-js."""

        def does_not_require_args():
            c = connection_from_array(array_abcde)
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def uses_default_connection_types():
            connection = connection_from_array(array_abcde[:1])
            assert isinstance(connection, Connection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == edge_a
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor=cursor_a,
                endCursor=cursor_a,
                hasPreviousPage=False,
                hasNextPage=False,
            )

        def accepts_custom_connection_type():
            class CustomConnection:
                # noinspection PyPep8Naming
                def __init__(self, edges, pageInfo):
                    self.edges = edges
                    self.page_info = pageInfo

            connection = connection_from_array(
                array_abcde[:1], connection_type=CustomConnection
            )
            assert isinstance(connection, CustomConnection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == edge_a
            page_info = connection.page_info
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor=cursor_a,
                endCursor=cursor_a,
                hasPreviousPage=False,
                hasNextPage=False,
            )

        def accepts_custom_edge_type():
            class CustomEdge:
                def __init__(self, node, cursor):
                    self.node = node
                    self.cursor = cursor

            connection = connection_from_array(array_abcde[:1], edge_type=CustomEdge)
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, CustomEdge)
            assert edge.node == "A"
            assert edge.cursor == cursor_a
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor=cursor_a,
                endCursor=cursor_a,
                hasPreviousPage=False,
                hasNextPage=False,
            )

        def accepts_custom_page_info_type():
            class CustomPageInfo:
                # noinspection PyPep8Naming
                def __init__(
                    self, startCursor, endCursor, hasPreviousPage, hasNextPage
                ):
                    self.startCursor = startCursor
                    self.endCursor = endCursor
                    self.hasPreviousPage = hasPreviousPage
                    self.hasNextPage = hasNextPage

            connection = connection_from_array(
                array_abcde[:1], page_info_type=CustomPageInfo
            )
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert edge == edge_a
            page_info = connection.pageInfo
            assert isinstance(page_info, CustomPageInfo)
            assert page_info.startCursor == cursor_a
            assert page_info.endCursor == cursor_a
            assert page_info.hasPreviousPage is False
            assert page_info.hasNextPage is False


def describe_connection_from_array_slice():
    def warns_for_deprecated_import():
        from importlib import reload

        with deprecated_call():
            from graphql_relay.connection import arrayconnection as deprecated

            # noinspection PyDeprecation
            reload(deprecated)
        # noinspection PyDeprecation
        assert deprecated.connection_from_array_slice is connection_from_array_slice

    def works_with_a_just_right_array_slice():
        c = connection_from_array_slice(
            array_abcde[1:3],
            dict(first=2, after=cursor_a),
            slice_start=1,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_b, edge_c],
            pageInfo=PageInfo(
                startCursor=cursor_b,
                endCursor=cursor_c,
                hasPreviousPage=False,
                hasNextPage=True,
            ),
        )

    def works_with_an_oversized_array_slice_left_side():
        c = connection_from_array_slice(
            array_abcde[0:3],
            dict(first=2, after=cursor_a),
            slice_start=0,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_b, edge_c],
            pageInfo=PageInfo(
                startCursor=cursor_b,
                endCursor=cursor_c,
                hasPreviousPage=False,
                hasNextPage=True,
            ),
        )

    def works_with_an_oversized_array_slice_right_side():
        c = connection_from_array_slice(
            array_abcde[2:4],
            dict(first=1, after=cursor_b),
            slice_start=2,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_c],
            pageInfo=PageInfo(
                startCursor=cursor_c,
                endCursor=cursor_c,
                hasPreviousPage=False,
                hasNextPage=True,
            ),
        )

    def works_with_an_oversized_array_slice_both_sides():
        c = connection_from_array_slice(
            array_abcde[1:4],
            dict(first=1, after=cursor_b),
            slice_start=1,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_c],
            pageInfo=PageInfo(
                startCursor=cursor_c,
                endCursor=cursor_c,
                hasPreviousPage=False,
                hasNextPage=True,
            ),
        )

    def works_with_an_undersized_array_slice_left_side():
        c = connection_from_array_slice(
            array_abcde[3:5],
            dict(first=3, after=cursor_b),
            slice_start=3,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_d, edge_e],
            pageInfo=PageInfo(
                startCursor=cursor_d,
                endCursor=cursor_e,
                hasPreviousPage=False,
                hasNextPage=False,
            ),
        )

    def works_with_an_undersized_array_slice_right_side():
        c = connection_from_array_slice(
            array_abcde[2:4],
            dict(first=3, after=cursor_b),
            slice_start=2,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_c, edge_d],
            pageInfo=PageInfo(
                startCursor=cursor_c,
                endCursor=cursor_d,
                hasPreviousPage=False,
                hasNextPage=True,
            ),
        )

    def works_with_an_undersized_array_slice_both_sides():
        c = connection_from_array_slice(
            array_abcde[3:4],
            dict(first=3, after=cursor_b),
            slice_start=3,
            array_length=5,
        )
        assert c == Connection(
            edges=[edge_d],
            pageInfo=PageInfo(
                startCursor=cursor_d,
                endCursor=cursor_d,
                hasPreviousPage=False,
                hasNextPage=True,
            ),
        )

    def describe_extended_functionality():
        """Test functionality that is not part of graphql-relay-js."""

        def does_not_require_args():
            c = connection_from_array_slice(array_abcde, slice_start=0, array_length=5)
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def uses_zero_as_default_for_slice_start():
            c = connection_from_array_slice(
                array_abcde[:1], dict(first=1), array_length=5
            )
            assert c == Connection(
                edges=[edge_a],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_a,
                    hasPreviousPage=False,
                    hasNextPage=True,
                ),
            )

        def uses_slice_end_as_default_for_array_length():
            c = connection_from_array_slice(
                array_abcde[:1], dict(first=1), slice_start=0
            )
            assert c == Connection(
                edges=[edge_a],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_a,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def ignores_len_of_slice_if_array_slice_length_provided():
            c = connection_from_array_slice(
                array_abcde[:2], dict(first=2), array_length=2, array_slice_length=1
            )
            assert c == Connection(
                edges=[edge_a],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_a,
                    hasPreviousPage=False,
                    hasNextPage=True,
                ),
            )

        def uses_array_slice_length_instead_of_len_function():
            class LettersWithoutLen:
                __getitem__ = array_abcde.__getitem__

            letters_without_len = cast(Sequence, LettersWithoutLen())

            with raises(TypeError):
                len(letters_without_len)

            with raises(TypeError):
                connection_from_array_slice(letters_without_len)

            c = connection_from_array_slice(letters_without_len, array_slice_length=5)
            assert c == Connection(
                edges=[edge_a, edge_b, edge_c, edge_d, edge_e],
                pageInfo=PageInfo(
                    startCursor=cursor_a,
                    endCursor=cursor_e,
                    hasPreviousPage=False,
                    hasNextPage=False,
                ),
            )

        def uses_default_connection_types():
            connection = connection_from_array_slice(
                array_abcde[:1], slice_start=0, array_length=1
            )
            assert isinstance(connection, Connection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == edge_a
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor=cursor_a,
                endCursor=cursor_a,
                hasPreviousPage=False,
                hasNextPage=False,
            )

        def accepts_custom_connection_type():
            class CustomConnection:
                # noinspection PyPep8Naming
                def __init__(self, edges, pageInfo):
                    self.edges = edges
                    self.page_info = pageInfo

            connection = connection_from_array_slice(
                array_abcde[:1],
                slice_start=0,
                array_length=1,
                connection_type=CustomConnection,
            )
            assert isinstance(connection, CustomConnection)
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert len(connection.edges) == 1
            assert edge == edge_a
            page_info = connection.page_info
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor=cursor_a,
                endCursor=cursor_a,
                hasPreviousPage=False,
                hasNextPage=False,
            )

        def accepts_custom_edge_type():
            class CustomEdge:
                def __init__(self, node, cursor):
                    self.node = node
                    self.cursor = cursor

            connection = connection_from_array_slice(
                array_abcde[:1], slice_start=0, array_length=1, edge_type=CustomEdge
            )
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, CustomEdge)
            assert edge.node == "A"
            assert edge.cursor == cursor_a
            page_info = connection.pageInfo
            assert isinstance(page_info, PageInfo)
            assert page_info == PageInfo(
                startCursor=cursor_a,
                endCursor=cursor_a,
                hasPreviousPage=False,
                hasNextPage=False,
            )

        def accepts_custom_page_info_type():
            class CustomPageInfo:
                # noinspection PyPep8Naming
                def __init__(
                    self, startCursor, endCursor, hasPreviousPage, hasNextPage
                ):
                    self.startCursor = startCursor
                    self.endCursor = endCursor
                    self.hasPreviousPage = hasPreviousPage
                    self.hasNextPage = hasNextPage

            connection = connection_from_array_slice(
                array_abcde[:1],
                slice_start=0,
                array_length=1,
                page_info_type=CustomPageInfo,
            )
            assert isinstance(connection, Connection)
            assert isinstance(connection.edges, list)
            assert len(connection.edges) == 1
            edge = connection.edges[0]
            assert isinstance(edge, Edge)
            assert edge == edge_a
            page_info = connection.pageInfo
            assert isinstance(page_info, CustomPageInfo)
            assert page_info.startCursor == cursor_a
            assert page_info.endCursor == cursor_a
            assert page_info.hasPreviousPage is False
            assert page_info.hasNextPage is False



import re
from typing import NamedTuple

__all__ = ["version", "version_info", "version_js", "version_info_js"]

version = "3.2.0"

version_js = "0.10.0"


_re_version = re.compile(r"(\d+)\.(\d+)\.(\d+)(\D*)(\d*)")


class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int

    @classmethod
    def from_str(cls, v: str) -> "VersionInfo":
        groups = _re_version.match(v).groups()  # type: ignore
        major, minor, micro = map(int, groups[:3])
        level = (groups[3] or "")[:1]
        if level == "a":
            level = "alpha"
        elif level == "b":
            level = "beta"
        elif level in ("c", "r"):
            level = "candidate"
        else:
            level = "final"
        serial = groups[4]
        serial = int(serial) if serial else 0
        return cls(major, minor, micro, level, serial)

    def __str__(self) -> str:
        v = f"{self.major}.{self.minor}.{self.micro}"
        level = self.releaselevel
        if level and level != "final":
            level = level[:1]
            if level == "c":
                level = "rc"
            v = f"{v}{level}{self.serial}"
        return v


version_info = VersionInfo.from_str(version)

version_info_js = VersionInfo.from_str(version_js)
